/* kolter_1616.c
 * 
 * Copyright (C) 1999 by Bernhard Kuhn <bkuhn@linux-magazin.de>
 * Copying Licence: GPL
 * Last Modification: Die Aug 17 11:29:37 CEST 1999
 *
 * history:
 * version 1.0beta1: initial version
 * version 1.0beta2: added openb/close for "module in use"-counter
 */



#include <linux/module.h>  /* generic module infos */
#include <linux/pci.h>     /* pci-bios functions   */
#include <linux/fs.h>      /* definitions for ioctl*/
#include <linux/delay.h>   /* udelay()             */
#include <asm/io.h>        /* inb(), outb() etc.   */
#include <asm/uaccess.h>   /* copy_to_user()       */





#include "kolter_1616.h"   /* ioctl-definitions   */





/* kolter 1616 general definitions */
#define PCI_VENDOR_ID_KOLTER             0x1001
#define PCI_DEVICE_ID_KOLTER_1616        0x0010
#define PCI_DEVICE_ID_KOLTER_1616_RELAIS 0x0013

/* kolter 1616 uses 8 io-port bytes */
#define KOLTER_1616_IO_EXTENT 0x08

/* Major number for /dev/kolter_1616.[0-7] */
#define KOLTER_1616_MAJOR 240

/* kolter 1616 register file */
#define KOLTER_1616_OUTPUT_A_ADDR   0x00
#define KOLTER_1616_OUTPUT_B_ADDR   0x01
#define KOLTER_1616_INPUT_A_ADDR    0x04
#define KOLTER_1616_INPUT_B_ADDR    0x05





/* the base addresses of the 1616-cards will stored here */
unsigned int kolter_1616_pci_ioaddr[KOLTER_1616_MAXBOARDS];

/* number of detected cards will be stored here */ 
int kolter_1616_pci_cards;





/* open-routine: only used to look if selected device is valid */
static int kolter_1616_open(struct inode *inode, struct file *file) {
  
  /* find out accessed device: /dev/kolter_1616.<minor> */
  unsigned int minor = MINOR(inode->i_rdev);
  
  
  /* ignore access to non-existent cards */
  if(minor>=kolter_1616_pci_cards) return -EIO;
 
  /* increase module usage counter */
  MOD_INC_USE_COUNT;

  return 0;

};





/* close-routine: only used to look if selected device is valid */
static int kolter_1616_release(struct inode *inode, struct file *file) {

  /* find out accessed device: /dev/kolter_1616.<minor> */
  unsigned int minor = MINOR(inode->i_rdev);

  
  /* ignore access to non-existent cards */
  if(minor>=kolter_1616_pci_cards) return -EIO;
 
  /* increase module usage counter */
  MOD_DEC_USE_COUNT;

  return 0;

};



/* ioctl-routine for input and output */
static int kolter_1616_ioctl(struct inode *inode, struct file *file,
			     unsigned int cmd, unsigned long arg) {
  
  /* find out accessed device: /dev/kolter_1616.<minor> */
  unsigned int minor = MINOR(inode->i_rdev);

  /* ignore access to non-existent cards */
  if(minor>=kolter_1616_pci_cards) return -EIO;
  
  /* process ioctl-command */
  switch(cmd) {
    
  case KOLTER_1616_OUTPUT:
    /* pass data to output */
    outw(arg,kolter_1616_pci_ioaddr[minor]+KOLTER_1616_OUTPUT_A_ADDR);  
    return 0;
    
  case KOLTER_1616_INPUT:
    /* get data from input */
    {
      unsigned int value=inw(kolter_1616_pci_ioaddr[minor]
			     +KOLTER_1616_INPUT_A_ADDR);
      copy_to_user((int*)arg,(int*)&value,sizeof(unsigned int));
    };
    return 0;
    
  };
  
  /* invalid ioctl-command */
  return -EIO;

};





/* what to do when specific file-operations occur */ 
static struct file_operations kolter_1616_fops = {
  NULL,  /* seek */
  NULL,  /* read */
  NULL,  /* write */
  NULL,  /* readdir */
  NULL,  /* poll */
  &kolter_1616_ioctl,  /* ioctl */
  NULL,  /* mmap */
  &kolter_1616_open,  /* open */
  NULL,  /* flush */
  &kolter_1616_release,  /* release */
  NULL,  /* fsync */
  NULL,  /* fasync */
  NULL,  /* check_media_change */
  NULL,  /* revalidate */
  NULL,  /* lock */
};





/* we may have to use deinitialisizing during initialisizing */
void cleanup_module(void);





/* initialisize module, called at 'insmod' */
int init_module(void) {

  /* reset number of found cards */
  kolter_1616_pci_cards=0;
  
  /* try to register major number for device access */
  if(register_chrdev(KOLTER_1616_MAJOR, "kolter_1616",
		     &kolter_1616_fops) == 0) {
    
    /* pci bios present ? */
    if(pcibios_present()) {
      
      unsigned char pci_bus,pci_device_fn;
      unsigned int pci_ioaddr;
      
      /* at least one Kolter 1616 present? */  
      while(pcibios_find_device(PCI_VENDOR_ID_KOLTER,
				PCI_DEVICE_ID_KOLTER_1616,
				kolter_1616_pci_cards,
				&pci_bus,&pci_device_fn) == 0) {
	
	/* more cards installed than supported? */
	if(kolter_1616_pci_cards==KOLTER_1616_MAXBOARDS) {
	  printk("kolter_1616: too many pci-cards present\n");
	  kolter_1616_pci_cards--;
	  cleanup_module();
	  return -EIO;
	};
	
	/* read base io-address */
	pcibios_read_config_dword(pci_bus,pci_device_fn,
				  PCI_BASE_ADDRESS_0,&pci_ioaddr);
	
	/* mask out io/mem-info since we know it is an io-address */
	pci_ioaddr &= PCI_BASE_ADDRESS_IO_MASK;
	
	/* normaly, regions should not yet be occupied by other cards,
	   so we can omit checking if regions are already reserved */
	request_region(pci_ioaddr,KOLTER_1616_IO_EXTENT,"kolter_1616");
	
	/* inform user */
	printk("kolter_1616: pci-card %i found at address 0x%x, ",
	       kolter_1616_pci_cards,pci_ioaddr);
	printk("bus=%i, device=%i\n",pci_bus,pci_device_fn>>3);
	
	/* store io-address of found pci-card */
	kolter_1616_pci_ioaddr[kolter_1616_pci_cards]=pci_ioaddr;
	
	/* try to find another kolter 1616 */
	kolter_1616_pci_cards++;
	
      };
      
      /* at least one Kolter 1616 with Relais present? */  
      while(pcibios_find_device(PCI_VENDOR_ID_KOLTER,
				PCI_DEVICE_ID_KOLTER_1616_RELAIS,
				kolter_1616_pci_cards,
				&pci_bus,&pci_device_fn) == 0) {
	
	/* more cards installed than supported? */
	if(kolter_1616_pci_cards==KOLTER_1616_MAXBOARDS) {
	  printk("kolter_1616: too many pci-cards present\n");
	  kolter_1616_pci_cards--;
	  cleanup_module();
	  return -EIO;
	};
	
	/* read base io-address */
	pcibios_read_config_dword(pci_bus,pci_device_fn,
				  PCI_BASE_ADDRESS_0,&pci_ioaddr);
	
	/* mask out io/mem-info since we know it is an io-address */
	pci_ioaddr &= PCI_BASE_ADDRESS_IO_MASK;
	
	/* normaly, regions should not yet be occupied by other cards,
	   so we can omit checking if regions are already reserved */
	request_region(pci_ioaddr,KOLTER_1616_IO_EXTENT,"kolter_1616_relais");
	
	/* inform user */
	printk("kolter_1616_relais: pci-card %i found at address 0x%x, ",
	       kolter_1616_pci_cards,pci_ioaddr);
	printk("bus=%i, device=%i\n",pci_bus,pci_device_fn>>3);
	
	/* store io-address of found pci-card */
	kolter_1616_pci_ioaddr[kolter_1616_pci_cards]=pci_ioaddr;
	
	/* try to find another kolter 1616 */
	kolter_1616_pci_cards++;
	
      };
      
      /* any kolter 1616 cards present? */
      if(kolter_1616_pci_cards) {

	printk("kolter_1616: %i pci-cards found\n",kolter_1616_pci_cards);
	return 0;
	
      };
      
      printk("kolter_1616: no pci-cards found\n");
      return -EIO;
      
    };
    
    printk("kolter_1616: no pci-bios present\n");
    return -EIO;
    
  };
  
  printk("kolter_1616: unable to get major %d\n",KOLTER_1616_MAJOR);
  return -EIO;

};





/* deinitialisize module, called at 'rmmod' */
void cleanup_module(void) {

  int index;

  /* unregister character device from defined major number */
  unregister_chrdev(KOLTER_1616_MAJOR,"kolter_1616");

  /* unregister io-regions from the kernel */
  for(index=0;index<kolter_1616_pci_cards;index++) {
    release_region(kolter_1616_pci_ioaddr[index],KOLTER_1616_IO_EXTENT);
  };

};

